#ifndef _RDBDATABASE_H_
#define _RDBDATABASE_H_


#include "RdbError.h"
#include "RdbSystem.h"
#include "List.h"
#include "LogHandler.h"

// forward declarations:
class RdbCursor;
class RdbCompilableDML;
class RdbQuery;
class RdbInsert;
class RdbDelete;
class RdbRelation;
class RdbJoin;
struct BitVector;


typedef PtrList<RdbCompilableDML> RdbCompilableDMLList;
typedef PtrList<RdbCursor> RdbCursorList;

class RdbDatabase
: public RdbError
{
  public:

    RdbDatabase(const String& newDbName,
		RdbSystem& newSystem,
		const String& newServerName);
				// attach to database newDbName using newSystem

    virtual ~RdbDatabase();

    ErrorCode Connect();	// make connect to this db
				// automatically called by queries, etc.

    ErrorCode Disconnect();	// disconnect from this db (db will commit)
				// automatically called by destructor

    ErrorCode ImmediateSql(const String& sql);
				// execute aribitrary sql
				// cannot need a cursor (no selects)

    // ErrorCode PrepareToCommit();

    ErrorCode Commit();		// commit the changes, release locks

    ErrorCode Rollback();	// undo changes, release locks

    String GetErrorMessage();	// get the error text

    RdbSystem& System() const;	// the system this db is in

    String DatabaseName() const; // Database name

    String ServerName() const;	// Server name

    String SystemName() const;	// System name

    virtual RdbQuery* NewQuery(RdbJoin& joinInfo, const BitVector& adorn) = 0;
				// allocate a new query for this
				// database & joinInfo

    virtual RdbInsert* NewInsert(RdbRelation& rel) = 0;
				// allocate a new insert statement for this
				// database & relation

    virtual RdbDelete* NewDelete(RdbRelation& rel, const BitVector& adorn) = 0;
				// allocate a new delete statement for this
				// database & relation

    ErrorCode SetFocus();	// automatically called by queries, etc.

    ErrorCode RemoveFocus();	// called by system on a switch

    virtual void const* Implementor() const = 0;

#if defined(DO_RDB_LOGGING)
    LogHandler& DBLog();	// where to send logging messages
#endif

  protected:

    RdbSystem& system;		// system this database is in

    const String dbName;	// name of the database to the backend

    const String serverName;	// name of the backend server

    RdbCursorList cursorList;	// cursors open in this database

    RdbCompilableDMLList compilableDMLList;
				// (possibly) prepared queries
				// used to be able to notify the queries
				// when they have been invalidated

    Boolean connected;		// are we connected to the backend?

#if defined(DO_RDB_LOGGING)
    LogHandler dblog;
#endif
    
    void InvalidateCompilableDMLs(); // mark all prepared queries as unprepared

    void InvalidateCursors();	// mark all queries as no longer valid

    // backend implementors:
    // See corresponding versions for more info.

    virtual ErrorCode VConnect() = 0; // establish connection to backend

    virtual ErrorCode VDisconnect() = 0; // disconnect from backend

    virtual ErrorCode VImmediateSql(const String& sql) = 0;
				// issue the statement now
				// this currently does not support any
				// statement that will produce a result
				// that will need to be printed

    // virtual ErrorCode VPrepareToCommit() = 0;

    virtual ErrorCode VCommit() = 0; // commit all changes

    virtual ErrorCode VRollback() = 0; // undo all changes

    virtual ErrorCode VPrepare(const String& id, const String& statement) = 0;
				// prepare the statement, identified by id
				// this is called by RdbCompilableDML::Prepare

    virtual String VGetErrorMessage() = 0;

    virtual ErrorCode VSetFocus() = 0; // reconnect if needed...

    virtual ErrorCode VRemoveFocus() = 0; // disconnect if you must...

  private:

    void RegisterCompilableDML( RdbCompilableDML& CompilableDML );  
				// called by CompilableDML constructor

    void DeregisterCompilableDML( RdbCompilableDML& CompilableDML ); 
				// called by CompilableDML destructor

    void RegisterCursor( RdbCursor& cursor );   
				// called by cursor constructor

    void DeregisterCursor( RdbCursor& cursor ); 
				// called by cursor destructor

    RdbDatabase(const RdbDatabase& other); // n/a

    RdbDatabase& operator =(const RdbDatabase& other); // n/a

    friend class RdbCompilableDML; // to register/deregister

    friend class RdbCursor;	// to register/deregister

};

#if defined(DO_RDB_LOGGING)
#   define RDBLOG(L,M) LOG_MSG(L,M)
#else
#   define RDBLOG(L,M)
#endif

#include "RdbDatabase.i"

#endif // _RDBDATABASE_H_
