#include <strstream.h>
#include "RdbSybaseQuery.h"
#include "RdbSybaseProcess.h"
#include "RdbSybaseDatabase.h"
#include "Log.h"


RdbSybaseQuery::RdbSybaseQuery(const RdbJoin& newJoin, 
			       const BitVector& newAdorn)
: RdbQuery(newJoin,newAdorn),
  database(*(RdbSybaseDatabase*)newJoin.Database().Implementor())
{ 
    CreateProcedure();
}


RdbSybaseQuery::~RdbSybaseQuery()
{
    RDBLOG(database.DBLog(), "Deleting query " << id << endl);
    database.ImmediateSql( String("drop procedure ") + id );
}


RdbError::ErrorCode RdbSybaseQuery::CreateProcedure()
{
    // create the statement without any conditions
    // begin the stmt to get type info
    // create the statement with variable conditions
    // create the procedure
    ErrorCode ec;
    id = System().GetOid();
    ostrstream s;
    s << "create procedure " << id;
    statement = CreateStmtWithoutConditions();
    StringList types;
    ec = GetTypes(types);
    if( ec )
      return Error(ec);
    const BitVector& adorn = Adornment();
    char sep = '(';
    for(int setBit = adorn.next(-1), v = 0;
	setBit >= 0 ;
	setBit = adorn.next(setBit), v++ ) {
	s << sep << " @p" << v << ' ' << types[setBit];
	sep = ',';
    }
    if( v > 0 ) s << ")";
    s << " as ";
    CreateStmt("@p", TRUE); // var id's = @p1, @p2, etc.
    s << statement;
    statement = s;
    ec = database.ImmediateSql(statement);
    if( ec )
      return Error(ec);
    return NO_ERROR;
}


RdbError::ErrorCode RdbSybaseQuery::GetTypes(StringList& typeList)
{
    assert(typeList.Size() == 0);
    RETCODE	rc;
    RdbSybaseProcess dbproc(database);
    if( dbproc == NULL ) {
	return Error(SQL_ERROR);
    }
    rc = dbcmd(dbproc, statement);
    if( rc != SUCCEED ) {
	RDBLOG(database.DBLog(), "failed to create dbcmd\n");
	return Error(SQL_ERROR);
    }
    rc = dbsqlexec(dbproc);
    if( rc != SUCCEED ) {
	RDBLOG(database.DBLog(), "failed to send command to sybase\n");
	return Error(SQL_ERROR);
    }
    rc = dbresults(dbproc);
    if( rc != SUCCEED ) {
	RDBLOG(database.DBLog(), "could not obtain results\n");
	return Error(SQL_ERROR);
    }
    int cols = dbnumcols(dbproc);
    for( int i = 1 ; i <= cols ; i++ ) {
	int type = dbcoltype(dbproc, i); 
	switch( type ) {
	  case SYBCHAR:
	  case SYBTEXT:
	  case SYBDATETIME:
	  case SYBDATETIME4:
	    int len = dbcollen(dbproc, i); 
	    typeList.Append(String("varchar(") + 
			    String(len) + 
			    String(")") );
	    break;

	  case SYBBIT:
	  case SYBINT1:
	  case SYBINT2:
	  case SYBINT4:
	    typeList.Append(String("int"));
	    break;

	  case SYBREAL:
	  case SYBFLT8:
	  case SYBMONEY:
	  case SYBMONEY4:
	    typeList.Append(String("float"));
	    break;

	    // case SYBBINARY:
	    // case SYBIMAGE:
	  default:
	    dbcanquery(dbproc);
	    return Error(UNSUPPORTED_TYPE);
	    
	}
	dbcanquery(dbproc);
    }
    return NO_ERROR;
}


RdbError::ErrorCode RdbSybaseQuery::VOpen(RdbCursor& cursor)
{
    RETCODE rc;

    RDBLOG(database.DBLog(), "Opening query " << id << '\n');

    SetCursorId(cursor, new RdbSybaseProcess(database));
    RdbSybaseProcess& dbproc = *(RdbSybaseProcess*)GetCursorId(cursor);

    rc = dbrpcinit(dbproc, id, 0);
    if( rc != SUCCEED )
      return Error(SQL_ERROR);

    const RdbItemList& bindings = cursor.Bindings();
    int cnt = bindings.Size();
    for( int i = 0 ; i < cnt ; i++ ) {
	int type;
	int len;
	int intVal;
	double doubleVal;
	void* data;
	const RdbItem& item = bindings[i];
	switch( bindings[i].Type() ) {
	  case RdbItem::RDB_STRING:
	    type = SYBCHAR;
	    len = String(item).Length();
	    data = (void*)(const char*)String(item);
	    break;
	  case RdbItem::RDB_INT:
	    type = SYBINT4;
	    len = -1;
	    intVal = item;
	    data = &intVal;
	    break;
	  case RdbItem::RDB_REAL:
	    type = SYBFLT8;
	    len = -1;
	    doubleVal = item;
	    data = &doubleVal;
	    break;
	  case RdbItem::RDB_NULL:
	    type = SYBCHAR;
	    len = 0;
	    data = NULL;
	    break;
	  default:
	    assert(0);		// unknown!!
	}
	rc = dbrpcparam(dbproc, NULL, 0, type, -1, len, (BYTE*)data);
	if( rc != SUCCEED )
	  return Error(SQL_ERROR);
    } // foreach binding

    rc = dbrpcsend(dbproc);
    if( rc != SUCCEED )
      return Error(SQL_ERROR);

    rc = dbsqlok(dbproc);
    if( rc != SUCCEED )
      return Error(SQL_ERROR);

    rc = dbresults(dbproc);
    if( rc != SUCCEED )
      return Error(SQL_ERROR);

    return NO_ERROR;
}


RdbError::ErrorCode RdbSybaseQuery::VClose(RdbCursor& cursor)
{
    RDBLOG(database.DBLog(), "Closing query " << id << '\n');
    RETCODE rc;
    RdbSybaseProcess& dbproc = *(RdbSybaseProcess*)GetCursorId(cursor);
    rc = dbcanquery(dbproc);
    if( rc != SUCCEED )
      return Error(SQL_ERROR);
    delete &dbproc;
    // how to get errors??  Ask database?? I want exception handling!!!
    return NO_ERROR;
}


RdbError::ErrorCode RdbSybaseQuery::VGetNext(RdbCursor& cursor,
					     RdbItemList& result)
{
    RDBLOG(database.DBLog(), "Get next on query " << id << '\n');
    RdbSybaseProcess& dbproc = *(RdbSybaseProcess*)GetCursorId(cursor);
    RETCODE rc;
    if( (rc = dbnextrow(dbproc)) == NO_MORE_ROWS ) {
	RDBLOG(database.DBLog(), "Last row returned\n");
	return SQL_EOF;
    }
    if( rc != REG_ROW ) {
	RDBLOG(database.DBLog(), "Error on get next\n");
	return Error(SQL_ERROR);
    }
    // Prepare result buffer:
    result.DeleteAll();
    int cols = Arity();
    for( int i = 1 ; i <= cols ; i++ ) {
	int len = dbdatlen(dbproc, i);
	if( len == 0 ) {	// column is null
	    result.Append( RdbItem() );
	} else {		// column is not null
	    int type = dbcoltype(dbproc, i);
	    BYTE* data = dbdata(dbproc, i);
	    switch( type ) {
	      case SYBCHAR:
	      case SYBTEXT:
		{
		    char* str = new char[len+1];
		    // -1 for destlen means trim spaces, & add null to the end.
		    if( dbconvert(NULL, type, data, len, SYBCHAR, 
				  (BYTE*)str, -1) <= 0 )
		      return Error(SQL_ERROR);
		    result.Append( RdbItem(String(&str)) );
				// str deleted by String
		}
		break;
	      case SYBDATETIME:
	      case SYBDATETIME4:
		{
		    const int datetimelen = 30;
		    char* str = new char[datetimelen+1];
		    int clen;
		    // -1 for destlen means trim spaces, & add null to the end.
		    if( (clen = dbconvert(NULL, type, data, len, SYBCHAR, 
					  (BYTE*)str, datetimelen)) <= 0 )
		      return Error(SQL_ERROR);
		    assert(clen <= datetimelen);
		    // trim & null terminate
		    if( dbconvert(NULL, SYBCHAR, (BYTE*)str, clen, SYBCHAR, 
				  (BYTE*)str, -1) <= 0 )
		      return Error(SQL_ERROR);
		    result.Append( RdbItem(String(&str)) );
				// str deleted by String
		}
		break;
	      case SYBBIT:
	      case SYBINT1:
	      case SYBINT2:
	      case SYBINT4:
		{
		    long l;
		    if( dbconvert(NULL, type, data, len, SYBINT4, (BYTE*)&l, 0)
		       <= 0 )
		      return Error(SQL_ERROR);
		    result.Append( RdbItem(l) );
		}
		break;
	      case SYBREAL:
	      case SYBFLT8:
	      case SYBMONEY:
	      case SYBMONEY4:
		{
		    double d;
		    if( dbconvert(NULL, type, data, len, SYBFLT8, (BYTE*)&d, 0)
		       <= 0 )
		      return Error(SQL_ERROR);
		    result.Append( RdbItem(d) );
		}
		break;
		// case SYBBINARY:
		// case SYBIMAGE:
	      default:
		assert(0);
	    } // switch
	} // else
    } // foreach column
    return NO_ERROR;
}


