/*****************************************************************************/
//    Arrays.C
/*****************************************************************************/

#include "Globals.h"
#include "BindEnv.h"
#include "BuiltinRelation.h"
#include "TupleIterator.h"
#include "Term.h"
#include "ArrayArg.h"
#include "LogicalArrayArg.h"
#include "FixedArrayArg.h"

#include "Interface.h"

/*****************************************************************************/

corBindEnv *corArraySolver(corBuiltinRelation& rel, corTupleIterator& iterator);
corBindEnv *corArrayDestructSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
corBindEnv *corArrayLookupSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
corBindEnv *corArrayBindSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
corBindEnv *corLogicalArrayBindSolver(corBuiltinRelation& rel, corTupleIterator& iterator);

/*****************************************************************************/
/*
    This file defines predicates that are used to manipulate the abstract
    data type array.  Note that the procedures here may perform
    _destructive_ update on the abstract data type.  In general this is
    _very_ dangerous, and may destroy the logical semantics of programs.
    The destructive update features here are an efficiency hack, and should
    not be used unless the user understands the CORAL system enough to figure
    out the operational implications of such updates.
*/
/*****************************************************************************/

/*****************************************************************************/

#define GET_ITERATOR_TERM( term, iterator, argnum, kind, subkind, msg) \
    if( corget_iterator_term( term, iterator, argnum, kind, subkind, msg)<0){\
        iterator.set_no_match();                    \
        return(NULL);                        \
    } 

/*****************************************************************************/
/*
    get_iterator_term is a utility routine to dereference an argument and 
    make sure that it is of a specified type / subtype
*/

int corget_iterator_term(corTerm& term, corTupleIterator &iterator, int argnum,
    arg_kind kind, int subkind, char *msg)
{
    term.bindenv = iterator.bindenv;
    term.expr = iterator.arg_list[argnum];
    corfull_dereference_term(term);
    
    if ((term.expr.kindof() != kind)
        ||
        (kind == COR_CONST_ARG &&
            ( subkind != ((corConstArg *)term.expr.other)->subkind()) )
        ||
        (kind == COR_NUM_CONST &&
            ( subkind != (int) (term.expr.cNum_kindof()))))
    {
        fprintf(corexEnv->error_file, "Error: %s: bad argument type to ", msg);
        term.printon(corexEnv->error_file);
        fputs("\n",corexEnv->error_file);
        return -1;
    }
    return 1;
}

/***************************************************************************/
/*
    Predicate Definitions:
    The following procedures define predicates that are used to create,
lookup, update and destroy objects of the abstract data type array.
    The arguments to the predicate are in iterator.  iterator.arg_list
is an ArgList data structure; variables in the arguments must be "dereferenced"
through iterator.bindenv.  Dereferencing is required since the arguments
of iterator have variables that have bindings specified in iterator.bindenv.
There are two forms of dereferencing -- a shallow dereferencing and a 
deep dereferencing.  The shallow form is done by 
    FULL_DEREFERENCE_TERM  
while the deep dereferencing is done by the method
    simplify  
simplify is defined in the file arg.C.  The two differ in that the first
dereferences an argument till its outermost level is not a variable, or is a
free variable.  There may be bound variables inside the structure.  The second
dereferences _all_ variables in the argument, and returns a structure where
the variables have been dereferenced, and any unbound variables have been 
renamed.
*/
/***************************************************************************/

/*********   
        The following lines show a method for declaring new predicates 
    that does not work currently.  It will hopefully be implemented 
    in later versions of CORAL.  For now, look at file builtin-rel.C
    to see how these predicates are declared.
BuiltinRelation dummy1 (1, EnterSymbol(ArrayDestructorString),
                 ArrayDestructSolver);
BuiltinRelation dummy2 (1, EnterSymbol(ArrayConstructorString), ArraySolver);
BuiltinRelation dummy2a (1, EnterSymbol(LogicalArrayConstructorString),
         LogicalArraySolver);
BuiltinRelation dummy3 (1, EnterSymbol(ArrayLookupString), ArrayLookupSolver);
BuiltinRelation dummy4 (1, EnterSymbol(ArrayBindString), ArrayBindSolver);
BuiltinRelation dummy4a (1, EnterSymbol(LogicalArrayBindString), 
        LogicalArrayBindSolver);
**********/

 
/***************************************************************************/
corBindEnv* corArraySolver(corBuiltinRelation& rel, corTupleIterator& iterator)
{

    // array(array,size)
    // Construct a new array of the specified size;  Default size = 16 for now.
    // Plan to change to growing arrays at some point.  Bind the first argument
    // to the newly constructed array.
    // If first arg is bound to array, finds size.

    corStackMark stackmark;

    int arity = iterator.arg_list.count();
    if (arity != 1 && arity != 2 ) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- bad number of arguments to %s :",
            corArrayConstructorString);
        iterator.arg_list.printon(corexEnv->error_file);
        fputs("\n",corexEnv->error_file);
        iterator.set_no_match();
        return(NULL);
    }

    corTerm term1 (iterator.arg_list[0], iterator.bindenv);
    FULL_DEREFERENCE_TERM(term1);
    if (term1.expr.kindof() != COR_VARIABLE ) { // Being used to find the size.

        GET_ITERATOR_TERM(term1, iterator, 0, COR_CONST_ARG, COR_ArrayKind,
             corArrayConstructorString);

        corTerm term2 (iterator.arg_list[1], iterator.bindenv);
        FULL_DEREFERENCE_TERM(term2);
    /***********
        if (term2.expr.kindof() != COR_VARIABLE ) {
            fprintf(exEnv->error_file,"CORAL :: Error -- bad argument to %s:",
                ArrayConstructorString);
            term22printon(exEnv->error_file);
            fputs("\n",exEnv->error_file);
            iterator.set_no_match();
            return(NULL);
        }
    *************/
        corTerm term3( corCArg (term1.expr.ArrayArgP()->size()), NULL);
        if (corunify_args(term2, term3) == COR_U_SUCCEED) {
            iterator.reset_no_match();
            return(iterator.bindenv) ;
        }
        stackmark.pop_to();
        iterator.set_no_match();
        return(NULL);
    }
    int size = 16;
    if (arity == 2) {
        corTerm term2;
        GET_ITERATOR_TERM(term2, iterator, 1, COR_NUM_CONST, COR_INTEGER,
            corArrayConstructorString);
        size = make_int(term2.expr);
    }
    
    corArg *newarray;

    if (!(rel.name->equals(corArrayConstructSymbol)))
        newarray = new corLogicalArrayArg(size);
    else
        newarray = new corFixedArrayArg(size);

    corunify_binding(term1.bindenv, ((corVarArg *)term1.expr.other)->var, newarray);
    iterator.reset_no_match();
    return(iterator.bindenv);
}

/***************************************************************************/
corBindEnv * corArrayDestructSolver(corBuiltinRelation&, corTupleIterator& iterator)
{ 
    // destruct(array):
    //   Destroy an array and deallocate space allocated for it.

    int arity = iterator.arg_list.count();
    if (arity != 1 ) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- bad number of arguments to %s:",
            corArrayDestructorString);
        iterator.arg_list.printon(corexEnv->error_file);
        fputs("\n",corexEnv->error_file);
        iterator.set_no_match();
        return(NULL);
    }

    corTerm term0;
    GET_ITERATOR_TERM(term0, iterator, 0, COR_CONST_ARG, COR_ArrayKind, 
        corArrayDestructorString); 
    delete term0.expr.other;
    iterator.reset_no_match();
    return(iterator.bindenv);
}

/***************************************************************************/
corBindEnv *corArrayLookupSolver(corBuiltinRelation&, corTupleIterator& iterator) 
{
    // lookup(array,i,val)
    //   Unify val with the i'th element of the array.  Array elements
    //   have a default value of [] (the nil list).
    //   Note unification will not affect the array elements since they
    //      are required to be ground.
    
    if (iterator.arg_list.count() != 3 ) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- bad number of arguments to %s:",
            corArrayLookupString);
        iterator.arg_list.printon(corexEnv->error_file);
        fputs("\n",corexEnv->error_file);
        iterator.set_no_match();
        return(NULL);
    }

    corTerm term0;
    GET_ITERATOR_TERM(term0, iterator, 0, COR_CONST_ARG, COR_ArrayKind,
        corArrayLookupString);

    corTerm term1;
    GET_ITERATOR_TERM(term1, iterator, 1, COR_NUM_CONST, COR_INTEGER,
        corArrayLookupString);

    int i = make_int(term1.expr);
    corCArg arg = term0.expr.ArrayArgP()->lookup(i);
    if (arg.IsEmpty()) {
        iterator.set_no_match();
        return(NULL);
    }

    corStackMark stackmark;
    corTerm term2(iterator.arg_list[2], iterator.bindenv);
    corTerm term3(arg, term0.bindenv);
    if (corunify_args(term2, term3) == COR_U_SUCCEED) {
        iterator.reset_no_match();
        return(iterator.bindenv) ;
    }
    stackmark.pop_to();
    iterator.set_no_match();
    return(NULL);
}

/***************************************************************************/
corBindEnv *corArrayBindSolver(corBuiltinRelation& , corTupleIterator& iterator) 
{
    // bind(array,i,val):
    //    bind array[i] to val.
    
    if (iterator.arg_list.count() != 3 ) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- bad number of arguments to %s:",
            corArrayBindString);
        iterator.arg_list.printon(corexEnv->error_file);
        fputs("\n",corexEnv->error_file);
        iterator.set_no_match();
        return(NULL);
    }

    corTerm term0;
    GET_ITERATOR_TERM(term0, iterator, 0, COR_CONST_ARG, COR_ArrayKind,
        corArrayBindString);

    corTerm term1;
    GET_ITERATOR_TERM(term1, iterator, 1, COR_NUM_CONST, COR_INTEGER,
        corArrayBindString);
    int i = make_int(term1.expr);
    
    // NOTE:: Should check that the arg. is a constant, and simplify it
    // before binding.  
    corTermLink *renamed_vars = NULL;
    corCArg newval = iterator.arg_list[2].simplify(iterator.bindenv,
        renamed_vars, NULL, NULL);
    if (renamed_vars != NULL) {
        fputs("Error: Terms containing variables cannot be stored in arrays!\n",
            corexEnv->error_file);
        fputs("\t - offending term = ",corexEnv->error_file);
        iterator.arg_list[2].print(iterator.bindenv, corexEnv->error_file);
        iterator.set_no_match();
        return(NULL);
    }

    term0.expr.ArrayArgP()->bind(i,&newval);
    iterator.reset_no_match();
    return(iterator.bindenv);
}

/***************************************************************************/
corBindEnv *corLogicalArrayBindSolver(corBuiltinRelation& , corTupleIterator& iterator)
{
    // logical_bind(array,i,val,newarray):
    //    create a version of array with array[i] bound to val, and assign the
    //    result to newarray.
    // Note that newarray must be a variable.

    if (iterator.arg_list.count() != 4 ) {
        fprintf(corexEnv->error_file,
            "CORAL :: Error -- bad number of arguments to %s:",
            corLogicalArrayBindString);
        iterator.arg_list.printon(corexEnv->error_file);
        fputs("\n",corexEnv->error_file);
        iterator.set_no_match();
        return(NULL);
    }

    corTerm term0;
    GET_ITERATOR_TERM(term0, iterator, 0, COR_CONST_ARG, COR_ArrayKind,
        corArrayBindString);

    corTerm term1;
    GET_ITERATOR_TERM(term1, iterator, 1, COR_NUM_CONST, COR_INTEGER,
        corArrayBindString);
    int i = make_int(term1.expr);

    corTerm newarray_term (iterator.arg_list[3], iterator.bindenv);
    FULL_DEREFERENCE_TERM(newarray_term);
    if (newarray_term.expr.kindof() != COR_VARIABLE ) {
        fprintf(corexEnv->error_file,"CORAL :: Error -- bad argument to %s:",
            corLogicalArrayBindString);
        newarray_term.printon(corexEnv->error_file);
        fputs("\n",corexEnv->error_file);
        iterator.set_no_match();
        return(NULL);
    }

    // NOTE:: Should check that the arg. is a constant, and simplify it
    // before binding.  
    corTermLink *renamed_vars = NULL;
    corCArg newval = iterator.arg_list[2].simplify(iterator.bindenv,
        renamed_vars, NULL, NULL);
    if (renamed_vars != NULL) {
        fputs("Error: terms containing variables cannot be stored in arrays!\n",
            corexEnv->error_file);
        fputs("\t - offending term = ",corexEnv->error_file);
        iterator.arg_list[2].print(iterator.bindenv, corexEnv->error_file);
        iterator.set_no_match();
        return(NULL);
    }

    corArrayArg *new_array = term0.expr.ArrayArgP()->make_version();
    // The value returned by make_version is a LogicalArrayArg or
    // FixedArrayArg depending on the type    of term0.expr. 
    new_array->bind(i,&newval);
    corTerm new_term(new_array,NULL);
    if (corunify_args(newarray_term, new_term) == COR_U_SUCCEED) {
        // Should succeed, since newarray_term is supposed to be a variable.
        iterator.reset_no_match();
        return(iterator.bindenv);
    }
    iterator.set_no_match();
    return(NULL);
}

/***************************************************************************/

