//**************************************************************************
//    Unparse.C
//**************************************************************************

#include "Unparse.h"

#include "Literal.h"
#include "Assert.h"
#include "BuiltinSymbols.h"
#include "Globals.h"

//**************************************************************************

static void print_unique_prednames(corClause *r, FILE *out);
static void sprint_adornment(corBitVector& adorn, char *str, int *pos);

//**************************************************************************

int corprint_new_line_after_rule=1;

/*****************************************************************************/
// print_adorn_for_pred is cleared when rules need to be displayed for tracing.
// At this stage, the predicate names already have the adornment.
// When the magic program calls the routine, the predicate names do not have 
// their adornments yet, and so the adornment needs to be printed out. 
//  The magic program uses the default setting of print_adorn_for_pred = 1.

int corprint_adorn_for_pred=1;
int corC_suppress_all_free=0;

//**************************************************************************
void corprint_rule(corClause *rule, FILE *out)
{
    int j;
    for (corClauseName *cname = rule->clause_names; cname; cname = cname->next) {
        //fprintf(out,"cname->name is '%s'\n", cname->name->string());

            // Need to add quotes if they aren't being used
        if (corexEnv->C_use_quotes)
            cname->name->print(0,out);
        else {
            fputs("\"",out);
            cname->name->print(0,out);
            fputs("\"",out);
        }
        if (cname->negated) fputs("-",out);
        fputs(": ",out);
    }
    corprint_pred(rule->head, out);
    for (int i=0; i < rule->head_deletes_count; i++) {
        fputs(", ",out);
        corprint_pred(rule->head_deletes[i], out, /* head */ 1);
    }
    if (rule->num_preds > 0 || rule->sips != NULL) {
        fputs(" :- \n\t\t",out);
        for (j = 0; j < rule->num_preds; ++j) {
            corprint_pred(rule->preds[j], out);
            if (j < rule->num_preds - 1)
            fputs( ", \n\t\t",out); // print a pred on a line.
        }
        if (j==0)
            fputs("true",out);    /* print "true" if no rhs preds */
        if (rule->sips) {
            fputs(" \n \t \t alias ",out);
            corprint_sips(rule, rule->sips, out);
        }
    }
    fputs(".",out);

    // Default is to print the new line. This is used by the magic program.
    // While dumping trace code, we dont want new lines printed. So then
    // the variable is cleared.
    if (corprint_new_line_after_rule)
        fputs("\n",out);
}

//**************************************************************************
void corprint_rule_with_order(corClause *rule, int *predOrders, int *markOffsets,
               int recursive, FILE* out)
{
    int j;

    corprint_pred(rule->head, out);
    for(int i=0; i < rule->head_deletes_count; i++) {
        fputs(", ",out);
        corprint_pred(rule->head_deletes[i], out, /* head */ 1);
    }
    if (rule->num_preds > 0 || rule->sips != NULL) {
        fputs(" :- \n\t\t",out);
        for (j = 0; j < rule->num_preds; ++j) {
            if ((markOffsets[2*predOrders[j]+1] % 2) == 0) // print old 
                fputs("old_",out);
            if ((j==0) && recursive) fputs("d_",out); // print delta
            corprint_pred(rule->preds[predOrders[j]], out);
            if (j < rule->num_preds - 1)
            fputs(", \n\t\t",out); // print a pred on a line.
        }

        if (j==0)
            fputs("true",out);    /* print "true" if no rhs preds */

//    if (rule->sips) {
//        fprintf(out, " \n \t \t alias ");
//        print_sips(rule, rule->sips, out);
//    }
    }
    fputs(".",out);

// Default is to print the new line. This is used by the magic program.
// While dumping trace code, we dont want new lines printed. So then
// the variable is cleared.
    if (corprint_new_line_after_rule)
         fputs("\n",out);
}

//**************************************************************************
static void sprint_pred( corLiteral *p, char *str, int *pos, int head=0)
{
    switch (p->kind) {
    case COR_REGULAR:
        if ( p->negated ) {
            if (head) 
                sprintf(str +*pos," del ");
            else 
                sprintf(str +*pos," not ");
            *pos += 5;
        }
        break;
    default:
        ASSERT((! p->negated ));
        break;
    }

    sprintf(str+(*pos), "%s", corSymbolString(p->pred));
    *pos += SymbolLength(p->pred);

    int temp = corC_suppress_all_free;
    corC_suppress_all_free = 1;
    sprint_adornment(p->adorn, str, pos);
    corC_suppress_all_free = temp;
    sprintf(str + *pos, "(");
    (*pos)++;
    p->args.sprint(str, pos);
    sprintf(str + *pos, ")");
    (*pos)++;
}

//**************************************************************************
void corsprint_rule(corClause *rule, char *str)
{
    int j;
    int str_pos = 0;

    for (corClauseName *cname = rule->clause_names; cname; cname = cname->next) {
        sprintf(str+str_pos, "%s", corSymbolString(cname->name));
        str_pos += SymbolLength(cname->name);
        if (cname->negated) {
            sprintf(str+str_pos, "-");
            str_pos++;
        }
        sprintf(str+str_pos, ": ");
        str_pos += 2;
    }

    sprint_pred(rule->head, str, &str_pos);
    for (j=0; j < rule->head_deletes_count; j++) {
        sprintf(str+str_pos, ", ");
        str_pos += 2;
        sprint_pred(rule->head_deletes[j], str, &str_pos, /* head */ 1);
    }
    if (rule->num_preds > 0 || rule->sips != NULL) {
        sprintf(str+str_pos, " :- ");
        str_pos += 4;
        for (j = 0; j < rule->num_preds; ++j) {
            sprint_pred(rule->preds[j], str, &str_pos);
            if (j < rule->num_preds - 1) {
                sprintf(str+str_pos, ", \n\t\t");
                str_pos += 5;
            }
        }
        if (j==0){
            sprintf(str+str_pos, "true");
            str_pos += 4;
        }

        // NOTE : Left out sips information
    }
    sprintf(str+str_pos, ".");
}

//**************************************************************************
/*
    Print a unique (within 'rule') name for the index'th predicate of 'rule',
    where index==0 means the head, and 1 is the first RHS predicate.
*/

static void corprint_unique_predname(corClause *rule, int index, FILE *out)
{
    corLiteral *pred = index ? rule->preds[index-1] : rule->head;

#    ifdef SHORT_SIP_PREDNAMES
        fprintf(out, "%c%d", corSymbolString(pred->pred)[0], index);
#    else
        char num_preds[12];
        sprintf(num_preds, "%d", rule->num_preds);
        fprintf(out, "%s", corSymbolString(pred->pred));
        // Letting the format depend on num_preds guarantees that the
        // same number of digits is used for all the preds.
        fprintf(out, "%0*d", (int) strlen(num_preds), index);
#    endif
}

//**************************************************************************
static void print_unique_prednames(corClause *r, FILE *out)
{
    int i;

    fputs("[",out);
    corprint_unique_predname(r, 0, out);
    for (i = 1; i <= r->num_preds; i++) {
        fputs(", ",out);
        corprint_unique_predname(r, i, out);
    }
    fputs("]",out);
}

//**************************************************************************
void corprint_pred( corLiteral *p, FILE *out, int head)
{
    switch (p->kind) {
    case COR_REGULAR:
        if ( p->negated ) {
            if (head) 
                fputs(" del ",out);
            else 
                fputs(" not ",out);
        }
        corPrintSymbol(p->pred, out);
        break;
    case COR_MAGIC:
        // CHANGE
        ASSERT((! p->negated ));
        fprintf(out, "m_%s", corSymbolString(p->pred));
        break;
    case COR_SUPMAGIC:
        // CHANGE
        ASSERT((! p->negated ));
        fprintf(out, "%s", corSymbolString(p->pred));
        break;
    case COR_FACTORMAGIC:
        ASSERT((! p->negated ));
        fprintf(out, "mf_%s", corSymbolString(p->pred));
        break;
    case COR_DONE_MAGIC:
        ASSERT((! p->negated ));
        fprintf(out, "done_m_%s", corSymbolString(p->pred));
        break;
    }

    int temp = corC_suppress_all_free;
    corC_suppress_all_free = 1;
    corprint_adornment(p->adorn, out);
    corC_suppress_all_free = temp;
    fputs("(",out);
    p->args.printon(out);
    fputs(")",out);
}

//**************************************************************************
void corprint_adornment_brack(corBitVector& a, FILE *file)
{
    int i;

    for (i = 0; i < a.len(); ++i) {
        fputs((a.test(i) ? "b" : "f"), file);
    }
}

//**************************************************************************
void corprint_sip(corClause *r, struct corsip *s, FILE *out)
{
    int corb;

    for (; s; s = s->next) {
        fputs("\n\t\t\t\t[",out);
        for (corb = s->tail.next(-1); corb >= 0; corb = s->tail.next(corb)) {
            corprint_unique_predname(r, corb, out);
            if (s->tail.next(corb) >= 0 ) 
                fputs(", ",out);
        }
        fputs("] ==> ",out);
        corprint_unique_predname(r, s->head, out);
        fputs("(",out);
        for (int c= s->bound_vars.next(-1),i=0; c>=0;
            c=s->bound_vars.next(c), i++)
        {
            fprintf(out, "%s", corSymbolString(r->var_names[c]));
            if ( i ) fprintf(out,",");
        }
        fputs(")",out);

        if (s->next)
            fputs(",\n",out);
        else
            fputs("\n",out);
    }
}

//**************************************************************************
void corprint_sips(corClause *r, struct corsips *s, FILE *out)
{
    int sip_printed = 0;
    print_unique_prednames(r, out);
    fputs("\n",out);
    if ( s ) fputs("\t\t sips [\n ",out);

    for (; s; s = s->next) {
        if( 1 /*s->sip_list*/ ) {
            if (sip_printed) fputs(",\n",out);
            sip_printed = 1;
            fputs("\t\t\t[ ",out);
            corprint_adornment_brack(s->adorn, out);
            fputs(": ",out);
            corprint_sip(r, s->sip_list, out);
            fputs("\t\t\t]",out);
        }
    }

    if ( !sip_printed )
        fputs("\t[ ]",out);
    else
        fputs("\t\t ]\n",out);
}

//**************************************************************************
static int all_args_free(corBitVector *a)
{
    return(allargsfree(a));
}

//**************************************************************************
static int all_args_free(char *a)
{
    for (int i=0; i < strlen(a); i++)
        if (a[i] == 'b')
            return 0;
    return 1;
}

//**************************************************************************
static int suppress_printing(corBitVector *a)
{
    if (! corprint_adorn_for_pred)
        return 1;

    if (a->len() == 0)
        return 1;

    if (corC_suppress_all_free && all_args_free(a))
        return 1;

    return 0;
}

//**************************************************************************
static int suppress_printing(char *a)
{
    if (! corprint_adorn_for_pred)
        return 1;
    if (strlen(a) == 0)
        return 1;
    if (corC_suppress_all_free && all_args_free(a))
        return 1;
    return 0;
}

//**************************************************************************
void sprint_adornment(char *buf, corBitVector* a)
{
    if (!a) return;

    if (! suppress_printing(a)) {
        *buf++ = '_';
        for (int i = 0; i < a->len(); i++)
            buf[i] = a->test(i) ? 'b' : 'f';
        buf[a->len()] = '\0';
    }
}

//**************************************************************************
void sprint_adornment(char *buf, char *a)
{
    if(!a) return;

    if ( ! suppress_printing(a)) {
        *buf++ = '_';
        sprintf(buf, "%s", a);
    }
}

//**************************************************************************
static void sprint_adornment(corBitVector& a, char *str, int *pos)
{
    if ( ! suppress_printing(&a)) {
        sprintf(str + *pos, "_"); (*pos)++;
        for (int i = 0; i < a.len(); ++i) {
            sprintf(str + *pos, a.test(i) ? "b" : "f");
            (*pos)++;
        }
    }
}

//**************************************************************************
void corprint_adornment(corBitVector& a, FILE *file)
{
    if (!suppress_printing(&a)) {
        fputs("_",file);
        for (int i = 0; i < a.len(); ++i)
            fputs((a.test(i) ? "b" : "f"), file);
    }
}

//**************************************************************************
void corprint_adornment(corBitVector *a, FILE *file)
{
    if (!a) return;
    corprint_adornment(*a, file);
}

//**************************************************************************
void corprint_adornment(char *a, FILE *file)
{
    if(!a) return;
    if ( !suppress_printing(a)) {
        fputs("_",file);
        fputs(a,file);
    }
}

//**************************************************************************
void corsprint_magic_name(char *buf, char *name, corBitVector *adorn)
{
    sprintf(buf, "m_%s", name);
    char *p = buf + strlen(buf);
    int temp = corC_suppress_all_free;
    corC_suppress_all_free = 1;
    sprint_adornment(p,adorn);
    corC_suppress_all_free = temp;
}

//**************************************************************************
void corsprint_magic_name(char *buf, char *name, char *adorn)
{
    sprintf(buf, "m_%s", name);
    char *p = buf + strlen(buf);
    int temp = corC_suppress_all_free;
    corC_suppress_all_free = 1;
    sprint_adornment(p,adorn);
    corC_suppress_all_free = temp;
}

//**************************************************************************
void corfprint_magic_name(FILE *out, char *name, corBitVector *adorn)
{
    fputs("m_",out);
    fputs(name,out);
    int temp = corC_suppress_all_free;
    corC_suppress_all_free = 1;
    corprint_adornment(adorn, out);
    corC_suppress_all_free = temp;
}

//**************************************************************************
void corfprint_magic_name(FILE *out, char *name, char *adorn)
{
    fputs("m_",out);
    fputs(name,out);
    int temp = corC_suppress_all_free;
    corC_suppress_all_free = 1;
    corprint_adornment(adorn, out);
    corC_suppress_all_free = temp;
}

//**************************************************************************
void corsprint_pred_name(char *buf, char *name, corBitVector *adorn)
{
    sprintf(buf, "%s", name);
    char *p = buf + strlen(buf);
    int temp = corC_suppress_all_free;
    corC_suppress_all_free = 1;
    sprint_adornment(p,adorn);
    corC_suppress_all_free = temp;
}

//**************************************************************************
void corsprint_pred_name(char *buf, char *name, char *adorn)
{
    sprintf(buf, "%s", name);
    char *p = buf + strlen(buf);
    int temp = corC_suppress_all_free;
    corC_suppress_all_free = 1;
    sprint_adornment(p,adorn);
    corC_suppress_all_free = temp;
}

//**************************************************************************
void corfprint_pred_name(FILE *out, char *name, corBitVector *adorn)
{
    fputs(name,out);
    int temp = corC_suppress_all_free;
    corC_suppress_all_free = 1;
    corprint_adornment(adorn, out);
    corC_suppress_all_free = temp;
}

//**************************************************************************
void corfprint_pred_name(FILE *out, char *name, char *adorn)
{
    fputs(name,out);
    int temp = corC_suppress_all_free;
    corC_suppress_all_free = 1;
    corprint_adornment(adorn, out);
    corC_suppress_all_free = temp;
}

//**************************************************************************
void corprint_allowed_adornment_annotations(corPredAnnotations *anno,FILE *file)
{
    corPredAnnotations *cur;

    for (cur = anno; cur; cur=cur->next()) {
        fprintf(file,"%s%s  ", "@",corSymbolString(corAllowedAdornSymbol));
        cur->print(file, 0); // adornment printing wrt non-adorned predicates
        fputs(".\n",file);
    }
}

//**************************************************************************
void corprint_make_index_annotations(corPredAnnotations *anno, FILE *file)
{

    corPredAnnotations *cur;

    for (cur = anno; cur; cur=cur->next()) {
        fprintf(file,"%s%s  ", "@",corSymbolString(corMakeIndexSymbol));
        cur->print(file);
        fputs(".\n",file);
    }
}

//**************************************************************************
void corprint_agg_sel_annotations(corPredAnnotations *anno, FILE *file)
{
    corPredAnnotations *cur;

    for (cur = anno; cur; cur=cur->next()) {
        fprintf(file,"%s%s  ", "@",corSymbolString(corAggSelectionSymbol));
        cur->print(file);
        fputs(".\n",file);
    }
}

//**************************************************************************
void corprint_diskrel_annotations(corPredAnnotations *anno, FILE *file)
{
    corPredAnnotations *cur;

    for (cur = anno; cur; cur=cur->next()) {
        fprintf(file,"%s%s  ", "@",corSymbolString(corDiskResidentSymbol));
        cur->print(file);
        fputs(".\n",file);
    }
}

//**************************************************************************
void corprint_prioritize_annotations(corPredAnnotations *anno, FILE *file)
{
    corPredAnnotations *cur;

    for (cur = anno; cur; cur=cur->next()) {
        fprintf(file,"%s%s  ", "@",corSymbolString(corPrioritizeSymbol));
        cur->print(file);
        fputs(".\n",file);
    }
}

//**************************************************************************
void corprint_MultisetAnnotations(corPredAnnotations *anno, FILE *file)
{
    corPredAnnotations *cur;
    
    for (cur = anno; cur; cur=cur->next()) {
        if (cur->check_subsum == 0) 
            fprintf(file,"%s%s    ", "@",corSymbolString(corMultisetSymbol));
        else
            fprintf(file,"%s%s    ", "@",corSymbolString(corCheckSubsumSymbol));
        
        cur->print(file, 0);
        fputs(".\n",file);
    }

}

//**************************************************************************
void corprint_Annotations(corModuleInfo& module_info, FILE *file,
    corPredAnnotations *index, corPredAnnotations *agg_sel,
    corPredAnnotations *allowed_adorn, corPredAnnotations *multiset,
    corPredAnnotations *prioritize_annotations, corPredAnnotations *diskrel)
{
    fprintf(file, "@%s %c.\n", corSymbolString(corNoPreProcessingSymbol),
        (module_info.PreProcessing ? '-' : '+'));
    fprintf(file, "@%s %c.\n", corSymbolString(corSupMagicIndexingSymbol),
        (module_info.SupMagicIndexing ? '+' : '-'));
    fprintf(file, "@%s %c.\n", corSymbolString(corSupplementaryMagicSymbol),
        (module_info.UseSupplementaryMagic ? '+' : '-'));
    fprintf(file, "@%s %c.\n", corSymbolString(corExistOptSymbol),
        (module_info.UseExistOpt ? '+' : '-'));
    fprintf(file, "@%s %c.\n", corSymbolString(corUseFactoringSymbol),
        (module_info.UseFactoring ? '+' : '-'));
    fprintf(file, "@%s %c.\n", corSymbolString(corOrdinaryMagicSymbol),
        (module_info.UseMagic ? '+' : '-'));
    fprintf(file, "@%s %c.\n", corSymbolString(corNoRewritingSymbol),
        (module_info.Rewriting ? '-' : '+'));
    fprintf(file, "@%s %c.\n", corSymbolString(corUseOrdSearchSymbol),
        (module_info.UseOrdSearch ? '+' : '-'));
    fprintf(file, "@%s %c.\n", corSymbolString(corSaveModuleSymbol),
        (module_info.SaveModule ? '+' : '-'));
    fprintf(file, "@%s %c.\n", corSymbolString(corUsePipeliningSymbol),
        (module_info.UsePipelining ? '+' : '-'));
    fprintf(file, "@%s %c.\n", corSymbolString(corSingleSccSymbol),
        (module_info.SingleScc ? '+' : '-'));
    fprintf(file, "@%s %c.\n", corSymbolString(corSingleAnswerSymbol),
        (module_info.SingleAnswer ? '+' : '-'));
    fprintf(file, "@%s %c.\n", corSymbolString(corNestedSccEvalSymbol),
        (module_info.NestedSccEval ? '+' : '-')); // added by Ken

    corprint_make_index_annotations(index, file);
    corprint_agg_sel_annotations(agg_sel, file);
    corprint_prioritize_annotations(prioritize_annotations, file);
    corprint_allowed_adornment_annotations(allowed_adorn, file);
    corprint_diskrel_annotations(diskrel, file);

    fprintf(file, "@%s %c.\n", corSymbolString(corCheckSubsumSymbol),
        (module_info.CheckSubsum ? '+' : '-'));
    fprintf(file, "@%s %c.\n", corSymbolString(corMultisetSymbol),
        (module_info.MultiSet ? '+' : '-'));
    corprint_MultisetAnnotations(multiset, file);


    fprintf(file, "@%s %c.\n", corSymbolString(corIndexDeltaSymbol),
        (module_info.IndexDeltas ? '+' : '-'));
    fprintf(file, "@%s %c.\n", corSymbolString(corReturnUnifySymbol),
        (module_info.UseReturnUnify ? '+' : '-'));
    fprintf(file, "@%s %c.\n", corSymbolString(corConvertFunctionsSymbol),
        (module_info.ConvertFunctions ? '+' : '-'));
    fprintf(file, "@%s %c.\n", corSymbolString(corNonGroundFactsSymbol),
        (module_info.NonGroundFacts ? '+' : '-'));
    fprintf(file, "@%s %c.\n", corSymbolString(corEagerEvalSymbol),
        (module_info.EagerEval ? '+' : '-'));
    fprintf(file, "@%s %c.\n", corSymbolString(corMonotonicSymbol),
        (module_info.Monotonic ? '+' : '-'));
    fprintf(file, "@%s %c.\n", corSymbolString(corPredicateSNSymbol),
        (module_info.UsePredicateSN ? '+' : '-'));
    fprintf(file, "@%s %c.\n", corSymbolString(corDiskResidentSymbol),
        (module_info.DiskRel ? '+' : '-'));
}

//**************************************************************************
