/*****************************************************************************/
//    Pipelined.C
/*****************************************************************************/

#include "Pipelined.h"

#include "QFModuleData.h"
#include "QFPipelinedModuleData.h"
#include "RuleInfo.h"
#include "CompileRule.h"
#include "ExportInfo.h"
#include "Literal.h"
#include "Compile.h"
#include "ParserStruct.h"
#include "DerivedMethod.h"

/*****************************************************************************/

static void init_module_preds(corQFModuleData *md);
static void generateIndices(corRuleInfo *rInfo, corQFModuleData *module);

/*****************************************************************************/
/*
    used to automatically generate make_index annotations for a rule.
    this function is similar to the identically named function in
    interp.C
*/

static void generateIndices(corRuleInfo *rInfo, corQFModuleData *module)
{
    int i;
    corVarLink *leading_vars, *indexing_vars ;

#    define RULEPTR rInfo->rule
#    define PREDINDEX(i) i

    //    The following code indexes base predicates in the rule
    for (i=0; i < RULEPTR->num_preds; i++) {
        // no point building an index on an arithmetic predicate
        if (!coris_a_builtin(RULEPTR->preds[PREDINDEX(i)]->pred)) {
            leading_vars = corvars_between(RULEPTR, 0, i-1, NULL, NULL);
            indexing_vars = corintersect_var_lists(
            corbuild_var_list(RULEPTR->preds[PREDINDEX(i)],NULL), leading_vars);
            corcreate_an_annotation(module->index_annotations,
                RULEPTR->preds[PREDINDEX(i)], indexing_vars);
        }
    }
#    undef RULE_STRUCT
#    undef PREDINDEX
}

/*****************************************************************************/
static void init_module_preds(corQFModuleData *md)
{
    corModulePred *predicates = md->predicates->next;

    for (; predicates; predicates = predicates->next) { 
        Name name = predicates->pred->name;
        int arity = predicates->pred->arity;
        predicates->offset = corget_offset(md->F, name, arity);
        switch (predicates->kind) {
        case QueryPredKind:
            predicates->check_subsum = cordetermine_subsum_status(md, name,
                md->module_info.CheckSubsum);
            break;
        case LocalPredKind:
            predicates->check_subsum = cordetermine_subsum_status(md, name,
                md->module_info.CheckSubsum);
            break;
        case MagicQueryPredKind:
            predicates->check_subsum = cordetermine_subsum_status(md, name,
                md->module_info.CheckSubsum);
            break;
        case ExternalPredKind:
            break;
        }
        corget_annos(md, predicates);
    }
}

/*****************************************************************************/
/* 
    This function is passed a pointer to the QFPipelinedModuleData 
    the information on the exported predicate.  This function basically
    does for a Pipelined QFModuleData what Scc analysis does for a
    Materialized one.
*/

void corinterpretPipelined(corQFPipelinedModuleData *md, corExportInfo *export_info,
    corParserStruct& corparserStruct)
{
    int i;
    corClause *tmp;
    int tvars,tliterals; // temporary vars used for calculating maximum vals

    corModulePred *predicates = md->predicates;

    Name exported_name = export_info->pred_name;
    corPredicate *pred = export_info->predicate; /* ptr to exported pred */
    
    md->numPreds = 0;
    for(corPredicate *p = corparserStruct.AllPredicates; p!= NULL; p = p->allprednext)
    {
        for (corClause *c = p->clauses;c != *p->clause_tail; c = c->pnext)
        {
            for (int ipred = 0; ipred < c->num_preds; ipred++) 
            {
                // not using Sccs for pipelining anyway
                c->preds[ipred]->predicate->work = NULL;
                c->preds[ipred]->predicate->scc = NULL;
        
                // know whether already added using add_pred_extern() or not
                c->preds[ipred]->predicate->mark = 0;
            }
        }
        md->numPreds++;
    }    
    
    // Make sure that all Relations used by the query are initialized
    // by initRulePointers. Use add_pred_extern to enter them into
    // the predicates collection to keep track of them.
    // First enter the the query predicate
    
    coradd_pred_extern(pred, QueryPredKind, predicates);
    // Have seen the exported predicate - all rules for it have been linked in
    pred->mark = 1;
    
    md->pred_ptrs = new corPredicate *[md->numPreds] ;
    md->indices = new int*[md->ruleArray->num_rules()];
    md->maxNumVars= new int[md->numPreds];
    md->maxNumLiterals = new int[md->numPreds];
    
    for (i = 0,p = corparserStruct.AllPredicates; p != NULL; p= p->allprednext,i++)
    {
        md->maxNumVars[i] = md->maxNumLiterals[i] = 0; 
        md->pred_ptrs[i] = p;
        for (corClause *c = p->clauses;c != NULL ; c = c->pnext) 
        {
            tliterals=((md->ruleArray)->rInfoPtrs[c->rule_number])->
                num_literals;
            tvars =((md->ruleArray)->rInfoPtrs[c->rule_number])->num_vars;
            // create indices
            md->indices[c->rule_number] = new int[((md->ruleArray)->
                rInfoPtrs[c->rule_number])->num_literals];
            if (tliterals > md->maxNumLiterals[i])
                md->maxNumLiterals[i] = tliterals;
            if (tvars > md->maxNumVars[i])
                md->maxNumVars[i] = tvars;

            // Iterate over all the Literals that are invoked in c
            for (int ipred = 0; ipred < c->num_preds; ipred++) 
            {
                corLiteral *child_literal = c->preds[ipred];
                corPredicate *child = child_literal->predicate;
                // If we have already seen this predicate then don't insert it 
                if (child->mark == 0)
                {
                    child->mark = 1;
                    // add this predicate too
                    tmp = child->clauses;
                    // If child->clauses == *child->clause_tail, then
                    // the child has no clauses, and is an ExternalPredKind.
                    coradd_pred_extern(child, tmp != *child->clause_tail ?
                        LocalPredKind : ExternalPredKind, predicates);
                }
            }
        }
    }
    
    md->clean_cache();
    md->export_pred = export_info->predicate;
    md->module_info = corparserStruct.CurModule;

    // move some of the run time work from initRulePointers to
    // compile time here. Determine what flags need to be set, etc.
    init_module_preds(md);
    
    // Insert the module structure along with its associated exported
    // query form into the DerivedMethod table.
    if (exported_name)
    {
        // The method really does not matter - not used anyway
        corDerivedMethod* query_method = new corDerivedMethod(exported_name,
            export_info->adornment,NULL , (corQFModuleData *)md);
    }
}

/*****************************************************************************/

