/*****************************************************************************/
//    ContextNode.C
/*****************************************************************************/

#include "Globals.h"
#include "ContextNode.h"

#include "StorageRelation.h"

/*****************************************************************************/
corContextNode::corContextNode(corGoalNode *new_goal, corContext *c) 
{
    // assuming that a context node is created with only one unmarked goal

    goals = new_goal;

    // set the ContextNode field of GoalNode 
    new_goal->c_node = this;

    // link in the new goal in the unmarked list
    corGoalNodeLink *g_link = new corGoalNodeLink(new_goal);
    first_unmarked_goal = g_link;
    last_unmarked_goal = g_link;

    first_marked_goal = NULL;
    last_marked_goal = NULL;

    mark = 0;                        // unmarked initially
    node_number = 0;

    negated = new_goal->negated;    // CHECK IF NEEDED??

    next = NULL;
    prev = NULL;
    next_marked = NULL;
    prev_marked = NULL;

    context = c;
}

/*****************************************************************************/
/*
ContextNode::~ContextNode()
{
free_all();
}
*/

/*****************************************************************************/
int corContextNode::any_unmarked_goals()
{
    // NOTE: we also have to check for deleted subgoals, which will be 
    // retained in the goals if the context node is
    // non-singleton because deleting nodes from a set is difficult

    // this also tries and moves nodes from the unmarked goals list to
    // the marked goals list so that space can be reclaimed later
    // NOTE: should not affect any asymptotics

    if (first_unmarked_goal == NULL) return 0;        // failure

    for (corGoalNodeLink *g_ptr = first_unmarked_goal; g_ptr != NULL; ) {
        if (!(g_ptr->g_node->is_deleted()))
            return 1;        // success        
        else {
            // the node pointed to by g_ptr is deleted
            // deallocate the node and update first_unmarked_goal
            first_unmarked_goal = g_ptr->next;
            move_to_marked_list(g_ptr);
            // NOTE: should not do a "delete g_ptr"
            // because that would result in unreclaimable space
        }
    }
    update_goalnode_info();
    return 0;
}

/*****************************************************************************/
corGoalNode* corContextNode::get_unmarked_goal()
{
    if (!any_unmarked_goals()) { 
        return NULL;
    }

    // choose the first goal node from unmarked list, mark it, 
    // update unmarked and marked lists
    corGoalNodeLink *g_ptr = first_unmarked_goal;
    first_unmarked_goal = first_unmarked_goal->next;
    move_to_marked_list(g_ptr);
    update_goalnode_info();

    corGoalNode *g = g_ptr->g_node;
    g->mark_goal();
        // marking context node and updating Context done from caller site
        // as is updating the node_number field of this node

    return g;
}

/*****************************************************************************/
void corContextNode::move_goals_to_done()
{
    ASSERT((is_marked()));
    ASSERT((first_unmarked_goal == NULL));

    // potentially there are two types of goals in the marked list
    // (1) magic facts: these have to be moved to the done relation
    // (2) supplementary facts: probably nothing has to be done to these

    for (corGoalNodeLink *g_tmp = first_marked_goal; g_tmp != NULL;
        g_tmp = g_tmp->next)
    {
        if ((g_tmp->g_node->is_magic_goal())&&(!(g_tmp->g_node->is_deleted())))
        {
            corStorageRelation *mp_done = g_tmp->g_node->find_done_relation();
#            ifdef DO_TRACE
                if (corexEnv->dbg_ordsearch) {
                    fprintf(corexEnv->trace_file,
                        "[Inserting %s(", corSymbolString(mp_done->name));
                    g_tmp->g_node->args().print(NULL,corexEnv->trace_file);
                    if (g_tmp->g_node->env_size == 0) 
                        fprintf(corexEnv->trace_file, ") - done (ground)]\n");
                    else 
                        fprintf(corexEnv->trace_file, ") - done (env-size:%d)]\n",
                                g_tmp->g_node->env_size);
                    context->print_all(corexEnv->trace_file);
                }
#            endif

            //    NOTE !! This is not a copy of the tuple, but the very same
            //    tuple.  Consequently, while emptying the done relation, the
            //    tuples should NOT be deleted.!!
            mp_done->insert_tuple(g_tmp->g_node);
                g_tmp->g_node->owner_rel = mp_done;
        }
    }
}

/*****************************************************************************/
void corContextNode::move_to_marked_list(corGoalNodeLink *g_ptr)
{
    if (first_marked_goal == NULL) {
        // create a new list
        first_marked_goal = g_ptr;
        last_marked_goal = g_ptr;
    }
    else {
        // insert at end of list
        last_marked_goal->next = g_ptr;
        last_marked_goal = g_ptr;
    }
    last_marked_goal->next = NULL;
}

/*****************************************************************************/
void corContextNode::merge_nodes(corContextNode *tbm)
{
    // the fields of tbm have to be merged into "this"

    // merge goals
    goals = goals->link_goals(tbm->goals);

    // merge unmarked goal lists
    if (first_unmarked_goal == NULL) {
        // nothing unmarked in "this"
        first_unmarked_goal = tbm->first_unmarked_goal;
        last_unmarked_goal = tbm->last_unmarked_goal;
    }
    else {
        if (tbm->first_unmarked_goal != NULL) {
            // something unmarked in tbm
            last_unmarked_goal->next = tbm->first_unmarked_goal;
            last_unmarked_goal = tbm->last_unmarked_goal;
        }
    }

    // merge marked goal lists
    ASSERT((first_marked_goal != NULL)); 
    if (tbm->first_marked_goal != NULL) {
        // something marked in tbm
        last_marked_goal->next = tbm->first_marked_goal;
        last_marked_goal = tbm->last_marked_goal;
    }

    // update node_number of the merged node to the higher of the two
    ASSERT((tbm->node_number > node_number));
    node_number = tbm->node_number;

    // NOTE: make sure caller in Context updates the first, last, etc. fields

    // NOTE: freeing up space of merged in node should be done from caller
}

/*****************************************************************************/
void corContextNode::update_goalnode_info()
{
    if (first_unmarked_goal == NULL) last_unmarked_goal = NULL;
    if (first_marked_goal == NULL) last_marked_goal = NULL;
}

/*****************************************************************************/
void corContextNode::free_all()
{
    // frees up the space of GoalNodeLink's but not the GoalNode's 
    // themselves because they are in a separate index and freeing
    // them would result in dangling pointers

    goals = NULL;

    corGoalNodeLink *g_ptr;
    for (g_ptr = first_unmarked_goal; g_ptr != NULL; ) {
        first_unmarked_goal = g_ptr->next;
        delete g_ptr;
        g_ptr = first_unmarked_goal;
    }
    last_unmarked_goal = NULL;

    for (g_ptr = first_marked_goal; g_ptr != NULL; ) {
        first_marked_goal = g_ptr->next;
        delete g_ptr;
        g_ptr = first_marked_goal;
    }
    last_marked_goal = NULL;

    mark = 0;
    node_number = 0;

    next = NULL;
    prev = NULL;
    next_marked = NULL;
    prev_marked = NULL;

    context = NULL;
}

/*****************************************************************************/
void corContextNode::print_all(FILE *file)
{
    fprintf(file, "\tPrinting ContextNode:\n");     
    fprintf(file, "\t\tMark = %d;  Node number = %d.\n", mark, node_number);

    corGoalNodeLink *temp;

    if (first_unmarked_goal == NULL) {
        fprintf(file, "\t\tNo unmarked goals.\n");
    }
    else {
        // there are unmarked goals
        fprintf(file, "\t\tList of unmarked goals =\n");
        for (temp = first_unmarked_goal; temp != last_unmarked_goal->next;
            temp = temp->next)
        {
            temp->g_node->print_all(file);
        }
    }

    if (first_marked_goal == NULL) {
        fprintf(file, "\t\tNo marked goals.\n");
    }
    else { // there are marked goals
        fprintf(file, "\t\tList of marked goals =\n");
        for (temp = first_marked_goal; temp != last_marked_goal->next;
                                        temp = temp->next) {
            temp->g_node->print_all(file);
        }
    }
    fprintf(file, "\n");
}

/*****************************************************************************/

