/*****************************************************************************/
//    StorageManager.C
/*****************************************************************************/

#ifdef WITH_PERSISTENCE



#include "Globals.h"
#include "CoralError.h"
#include "SMUtil.h"
#include "sm_client.h"
#include "Config.h"
#include "Schema.h"
#include "IndexStruct.h"
#include "BPlusTreeIndex.h"
#include "FileDescStruct.h"
#include "BufferGroup.h"
#include "Transaction.h"
#include "Catalog.h"
#include "StorageManager.h"


#include "BuiltinRelation.h"


/*****************************************************************************/

#define DATA_VOL 8000
#define TEMP_VOL 7000

/*****************************************************************************/

corStorageManager SM;
int verboseError;

// NEW_CHNG:wX
int SM_initialized = 0;

extern corBindEnv *SchemaSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
extern corBindEnv *DelSchemaSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
extern corBindEnv *StartBulkLoadSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
extern corBindEnv *EndBulkLoadSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
extern corBindEnv *TransBeginSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
extern corBindEnv *TransAbortSolver(corBuiltinRelation& rel, corTupleIterator& iterator);
extern corBindEnv *TransCommitSolver(corBuiltinRelation& rel, corTupleIterator& iterator);

extern void propagateChangesToDisk();
/*****************************************************************************/
void corStorageManager::init()
{
    int e;

    SM_initialized = 0;
    verboseError = !corexEnv->C_quiet_mode_default;

    if ((corexEnv->C_exec_mode == COR_DOING_FACTOR_MAGIC) ||
        (corexEnv->C_exec_mode == COR_DOING_MAGIC) ||
        (corexEnv->C_exec_mode == COR_DOING_FACTOR) ||
        (corexEnv->C_exec_mode == COR_DOING_EXIST))
            return;

    //    Initialize the storage manager

    char *error_msg;
    int bufsize = -1;

    e = sm_ReadConfigFile(NULL, NULL, &error_msg);
    if (e >= 0) {
        e = sm_GetClientOption("bufpages", (void *)(&bufsize));
        if (ErrorCheck(e, "sm_GetClientOption")) return;
    }

    if (bufsize < 0 ) {
        // Use default config values
        bufsize = COR_CLIENT_BUF_SIZE;
        e = sm_SetClientOption("bufpages", (void *)(&bufsize), SM_int);
        if (ErrorCheck(e, "sm_SetClientOption")) return;
    }

    e = sm_Initialize();

    if (ErrorCheck(e, "sm_Initialize")) return;

    //    Get the disk name
    char* volidName = getenv("EVOLID");
    // Default disk name is 'coraldisk'
    if (!volidName) _volid = DATA_VOL;
    else _volid = (VOLID)(atoi(volidName));

    char* tempvolName = getenv("ETEMPVOLID");
    if (!tempvolName) _temp_volid = TEMP_VOL;
    else _temp_volid = (VOLID)(atoi(tempvolName));

    //    Mount the volumes.
    e = sm_MountVolume(_volid);
    if (ErrorCheck(e, "sm_MountVolume")) {
        fprintf(corexEnv->error_file, "Volid of volume : %d\n", _volid);
        return;
    }
    e = sm_MountVolume(_temp_volid);
    if (ErrorCheck(e, "sm_MountVolume")) {
        fprintf(corexEnv->error_file, "Volid of volume : %d\n", _temp_volid);
        return;
    }

    //    Open a buffer group.
    if (buf.open(COR_SM_BUF_SIZE, BF_MRU)) return;

    SM_initialized = 1;
    
    startTrans();

    if (catalog.init("CORAL_CATALOG_FILE", "CORAL_CATALOG_INDEX")) {
        SM_initialized = 0;
        return;
    }

    commitTrans();

    //    Install schema, start_bulk_load and end_bulk_load builtin relations

    corRelation *temp;
    temp = new corBuiltinRelation(1, corEnterSymbol("schema"), SchemaSolver);
    temp = new corBuiltinRelation(1, corEnterSymbol("del_schema"), DelSchemaSolver);
                                                                             
    temp = new corBuiltinRelation(1, corEnterSymbol("start_bulk_load"),
        StartBulkLoadSolver);
    temp = new corBuiltinRelation(1, corEnterSymbol("end_bulk_load"),
        EndBulkLoadSolver);

    temp = new corBuiltinRelation(0, corEnterSymbol("t_begin"), TransBeginSolver);
    temp = new corBuiltinRelation(0, corEnterSymbol("t_commit"), TransCommitSolver);
    temp = new corBuiltinRelation(0, corEnterSymbol("t_abort"), TransAbortSolver);
}

/*****************************************************************************/
void corStorageManager::shutdown()
{
    if (SM_initialized) {
        startTrans();
        propagateChangesToDisk();
        commitTrans();

        sm_AbortTransaction(trans.tid);
        sm_CloseBufferGroup(buf.bufGroup);
        sm_DismountVolume(_volid);
        sm_ShutDown();
        SM_initialized = 0;
    }
}

/*****************************************************************************/
/*
    Retrieves an object from the storage manager.
    The var parameter is a pointer which, when the function returns,
    points to the retrieved object in the buffer pool. 

    The retrieved object is pinned in the buffer pool
    and will have to be explicitly unpinned using an sm_release() call.
    The object should be accessed only INDIRECTLY via the 'basePtr'
    field of the USERDESC structure, because EXODUS may relocate the
    contents of the buffer pool. Also note that the object is now
    PINNED in the buffer pool, and has to be unpinned using the function
    releaseObject().
*/

int corStorageManager::retrieveObject(OID *oid, USERDESC ** u_desc)
{
    if (!SM_initialized) return 0;

    int e = sm_ReadObject(bufGroup(), oid, 0, READ_ALL, u_desc);
    if (ErrorCheck(e, "retrieveObject:sm_ReadObject")) return 0;

    return 1;
}

/*****************************************************************************/
//    Releases an object pinned in the buffer pool.

int corStorageManager::releaseObject(USERDESC *u_desc)
{
    if (!SM_initialized) return 0;

    int e = sm_ReleaseObject(u_desc);
    if (ErrorCheck(e, "releaseObject:sm_ReleaseObject")) return 0;

    return 1;
}

/*****************************************************************************/
//    Destroys an object. If bufGroup is 0, the object is in the catalog file.

int corStorageManager::destroyObject(OID *oid, int bufGroup)
{
    if (!SM_initialized) return 0;

    int e;

    // If bufGroup is 0, the object belongs to the catalog
    if (!bufGroup) bufGroup = this->bufGroup();

    e = sm_DestroyObject(bufGroup, oid);
    if (ErrorCheck(e, "destroyObject:sm_DestroyObject")) return 0;

    return 1;
}

/*****************************************************************************/
int corStorageManager::lookupFile(Name name, int arity, OID* oid)
{
    if (!SM_initialized) return 0;

    char buf[COR_MAX_REL_NAME_LEN + 6];

    if (arity > COR_MAX_ARGLIST_SIZE) {
        CORAL_error(COR_TOO_MANY_ARGS, NULL, "lookupFile");
        return 0;
    }

    ASSERT(name->length() <= COR_MAX_REL_NAME_LEN);
    sprintf(buf, "%s%3d", name->string(), arity);

    return catalog.lookupFile(buf, oid);
}

/*****************************************************************************/
int corStorageManager::addFile(Name name, int arity, corFileDescStruct *obj)
{
    if (!SM_initialized) return 0;

    char buf[COR_MAX_REL_NAME_LEN + 6];

    if (arity > COR_MAX_ARGLIST_SIZE) {
        CORAL_error(COR_TOO_MANY_ARGS, NULL, "addFile");
        return 0;
    }

    ASSERT(name->length() <= COR_MAX_REL_NAME_LEN);
    sprintf(buf, "%s%3d", name->string(), arity);

    return catalog.addFile(buf, obj);
}

/*****************************************************************************/
int corStorageManager::deleteFile(Name name, int arity)
{
    if (!SM_initialized) return 0;

    OID oid;
    USERDESC *u_desc;

    // Look up the file in the catalog
    if (! this->lookupFile(name, arity, &oid)) return 0;

    // Retrieve the file descriptor object
    if (! this->retrieveObject(&oid, &u_desc)) return 0;
    corFileDescStruct fd;
    fd.copyDesc((corFileDescStruct *)(u_desc->basePtr));

    // Destroy the file and its contents
    int e = sm_DestroyFile(SM.bufGroup(), &(fd.fid));
    if (ErrorCheck(e, "sm_DestroyFile")) return 0;

    // Destroy all indexes on the file
    for (int i = 0; i < fd.num_indexes; i++) {
        if (fd.indexes[i].valid()) {
            e = sm_DestroyIndex(&(fd.indexes[i].iid), SM.bufGroup());
            if (ErrorCheck(e, "sm_DestroyIndex")) return 0;
        }
    }

    // Delete the entry for the file in the catalogs
    char buf[COR_MAX_REL_NAME_LEN + 6];

    if (arity > COR_MAX_ARGLIST_SIZE) {
        CORAL_error(COR_TOO_MANY_ARGS, NULL, "deleteFile");
        return 0;
    }

    ASSERT(name->length() <= COR_MAX_REL_NAME_LEN);
    sprintf(buf, "%s%3d", name->string(), arity);

    if (!catalog.deleteFile(buf, &oid)) return 0;
        
    this->releaseObject(u_desc);
    return 1;
}

/*****************************************************************************/
void corStorageManager::listFiles(FILE *outf)
{
    if (!SM_initialized) return;

    catalog.listFiles(outf);
}

/*****************************************************************************/

#endif WITH_PERSISTENCE

