
//**************************************************************************
//    ConnectArray.C
//
//    The ConnectArray Class - Used to hold an array of connections. This
//    class is the top-level item in the implemenatation of the Coral
//    client/server stuff. A ConnectArray holds an array of Connections,
//    which each hold an array of Users.  The ConnectArray also contains
//    a CoralSock object for the server port.  To initialize, EITHER
//    BindToPort() (for server-mode) OR LocalConnection() (for a single
//    stdin/stdout connection) should be called to initialize the
//    ConnectArray object.
//     
//**************************************************************************

#include "Identify.h"
#include "ConnectArray.h"
#include "Globals.h"
#include "ClientProtocol.h"
#include "Assert.h"

//**************************************************************************
// ConnectArray Constructor - Just clears everything.
//**************************************************************************
corConnectArray::corConnectArray()
{
    nxt_conn = 0;
    max_conn = MAX_CONNECTIONS;

    for (int i = 0; i < MAX_CONNECTIONS; i++)
        connArray[i] = (corConnection *) NULL;

    memset(&select_msk, '\0', sizeof(select_msk));

    serv_sock = (CoralSock *) NULL;
#ifdef    DO_EXPLAIN
    expl_sock = (CoralSock *) NULL;
#endif
}

//**************************************************************************
// ConnectArray Destructor - Closes the server socket (if any), then closes
//    each connection in the array.
//**************************************************************************
corConnectArray::~corConnectArray()
{
      // Close the server port if not local connection
    if (serv_sock && !(serv_sock->Local()))
        delete serv_sock;

#ifdef    DO_EXPLAIN
      // Close explain server socket
    if (expl_sock)
        delete expl_sock;
#endif

      // Close all connections
    for (int i = 0; i < MAX_CONNECTIONS; i++)
        if (connArray[i])
            delete connArray[i];
}

//**************************************************************************
// BindToPort() -
//    This call initializes the ConnectArray object as a server, the port
//    given is the "well-known" port to bind to, and accept connections on.
//
//**************************************************************************
int corConnectArray::BindToPort(unsigned short port)
{
    if (serv_sock)
        return(0);

    serv_sock = new CoralSock(port);
    if (serv_sock->socketfd < 0)
        return(-1);

    FD_SET(serv_sock->socketfd, &select_msk);

#ifdef    DO_EXPLAIN

      // Explain server socket
    unsigned short eport = EXPLAIN_PORT;
    expl_sock = new CoralSock(eport);

    if (expl_sock->socketfd < 0){
        delete serv_sock;
        return(0);
    }

    FD_SET(expl_sock->socketfd, &select_msk);
#endif

    // This next message is helpful for the test suite programs.
    fprintf(corexEnv->print_file,
            "[%d]: Coral Server started on port %hu.\n", getpid(), port);
    fflush(corexEnv->print_file);
    return(1);
}

//**************************************************************************
// LocalConnection() -
//    This call initializes the ConnectArray object NOT in server mode.
//    It sets up a "local" connection (to stdin/stdout) that looks/acts
//    as much like a true server connection as possible.
//    NOTE: NextConn() should not be called if the ConnectArray has been
//    initialized using LocalConnection().
//**************************************************************************
int corConnectArray::LocalConnection(int use_stdin)
{
      // If we are already a server, cannot change to non-server mode.
    if (serv_sock) 
        return(0);

      // 0 specifies to use stdin/stdout
    serv_sock = new CoralSock((unsigned short) 0);

      // Protection in case NextConn() is foolishly called while not
      // in server mode.
    FD_SET(0,&select_msk);

      // Place to start looking for next connection
    connArray[0] = new corConnection(serv_sock, use_stdin);
    connArray[0]->conn_id = 0;

    num_conn = 1;

      // Start using the connection
    connArray[0]->useConnection();

    return(1);
}

//**************************************************************************
// NextConn() -
//    This function finds the next available connection with data pending
//    and switches to it.  This function also checks for connection
//    requests on the server socket and accepts them. If more than one
//    connection has data waiting to be read, repeatedly calling this
//    function will switch between them in a round-robin fashion.
//
//    Return: The return value is the number of connections with data,
//        or < 0 if an error occurs.
//
//**************************************************************************
int corConnectArray::NextConn()
{
    int    num = 0;
    fd_set tmp_msk;

    do {
          // Duplicate the select mask.
        memcpy(&tmp_msk, &select_msk, sizeof(fd_set));


        // Find next connection (round-robin) with data pending
        // Block until we recieve a request to connect or data
        // If a connection request is recieved, connect it and
        // continue waiting for data on any port.

#ifdef HPUX
        if ((num = select(FD_SETSIZE,(int *) &tmp_msk,0,0,
                            (struct timeval *) NULL)) <= 0){
#else
        if ((num = select(FD_SETSIZE,&tmp_msk,0,0,
                            (struct timeval *) NULL)) <= 0){
#endif
            perror("Coral: Select Error ");
            exit(-1);
        }
    
        if (FD_ISSET(serv_sock->socketfd, &tmp_msk)){
    
              // Connection request 
            NewConnection();
            num--;
        }

#ifdef    DO_EXPLAIN

        if (FD_ISSET(expl_sock->socketfd, &tmp_msk)){

              // Connection request (for Explain)
            ExplainConnection();
            num--;
        }
#endif

    } while(!num);

      // If only 1 port has data, start at the one you're currently at.
      // Otherwise, start at the next (round-robin)
    if (num != 1)
        cur_conn = (cur_conn + 1) % MAX_CONNECTIONS;

    int prv_conn = cur_conn;

    do {

        corConnection *conn;

        if ((conn = connArray[cur_conn]) 
                && FD_ISSET(conn->conn_sock->socketfd, &tmp_msk)){

              // Make this connection the current
            conn->useConnection();

              // Return the number of connections with data (why not?)
            return(num);
        }

        cur_conn = (cur_conn + 1) % MAX_CONNECTIONS;

    } while(cur_conn != prv_conn);

    fprintf(stderr, "Could not locate a socket with data?!\n");

    return(-1);
}

//**************************************************************************
// NewConnection() -
//    This function is callled when a connection request is recieved on the
//    server port.  This function decides if the new connection should be
//    allowed and accepts or refuses it. 
//
//    Return: The return value is the connection id of the new connection,
//        or < 0 if an error occurs or the connection is refused.
//
//**************************************************************************
int corConnectArray::NewConnection()
{
    int this_conn;

    if (!serv_sock) 
        return(-1);

      // Check for at max connections
    if (num_conn >= max_conn){
        serv_sock->RefuseConnection();
        return(-1);
    }

      // The max_conn value can be changed dynamically
    if (nxt_conn >= max_conn)
        nxt_conn = 0;

    this_conn = nxt_conn;
    while(connArray[this_conn] && (this_conn < max_conn))
        this_conn++;

      // Place to start looking for next connection
    nxt_conn = this_conn + 1; 
    connArray[this_conn] = new corConnection(serv_sock);
    connArray[this_conn]->conn_id = this_conn;

    FD_SET((connArray[this_conn])->conn_sock->socketfd, &select_msk);
    num_conn++;

    return(this_conn);
}

//**************************************************************************
// DeleteConnection() -
//    This function deletes a connection from the ConnectArray object.
//
//**************************************************************************
void corConnectArray::DeleteConnection(int conn_indx)
{
    if (! connArray[conn_indx])
        return;

    FD_CLR((connArray[conn_indx])->conn_sock->socketfd, &select_msk);

    delete connArray[conn_indx];
    connArray[conn_indx] = (corConnection *) NULL;

    num_conn--;

      // Keep track of lowest available connection
    nxt_conn = (conn_indx < nxt_conn) ? conn_indx : nxt_conn;
}

//**************************************************************************
// ExplainConnection() -
//    This function is callled when a connection request is recieved on the
//    explain port.  This function decides if the connection attempt is
//    valid and accepts or refuses it.
//
//    Return: The return value is zero if everything is okay
//        or < 0 if an error occurs or the connection is refused.
//
//**************************************************************************
int corConnectArray::ExplainConnection()
{
#ifdef    DO_EXPLAIN

    char buff[500];
    int c_id=0, u_id=0, pnt=0;

    if (!expl_sock) 
        return(-1);

    CoralSock *tmp_sock = expl_sock->AcceptConnection();
    FILE *in = tmp_sock->InFilePtr();

    if (!fgets(buff, 500, in)){
        return(-1);
    }

    sscanf(buff, "%d %d %n", &c_id, &u_id, &pnt); 
   
    if (u_id){ 
        for (int i=pnt; buff[i]; i++)
            if ((buff[i] == '\r') || (buff[i] == '\n'))
                buff[i] = '\0';
    } else
        buff[pnt] = '\0';

    corConnection *conn = connArray[c_id];
    corUser *user;

    if ((! conn) || (! (user = conn->userArray[u_id])) ||
            strcmp(user->password,&buff[pnt]) || 
            (user->corexEnv.explain_fp != NULL) || 
            ! (user->corexEnv.GlobalRelationOptions & REL_EXPLAIN)){

        int sd = tmp_sock->socketfd;
        char *mess = "Connection refused by Coral Server.\n";
        write(sd,mess,strlen(mess));

          // Disconnect
        shutdown(sd,2);
        close(sd);
        return(0);
    }

      // They check out, set them up to recieve explain info
    user->explain_sock = tmp_sock;
    user->corexEnv.explain_fp = tmp_sock->OutFilePtr();
      // And inform client (client blocks waiting for a message)
    fprintf(user->corexEnv.print_file, "%sExplain Connected.\n", CP_MESG);
    fflush(user->corexEnv.print_file);
    fputs("coral.\n",user->corexEnv.explain_fp);

#endif    // DO_EXPLAIN

    return(0);
}

//**************************************************************************
// ListWho() - 
//    This function lists all users for the connection specified, 
//    (printing the output to the file pointer given), if the connection
//    specified is NULL (the default value), lists all users on all
//    connections.
//
//    Return: The number of users on this (or all) connections.
//
//**************************************************************************
int corConnectArray::ListWho(FILE *fp, corConnection *conn)
{
    int cnt = 0;

    if (conn){
        cnt += conn->ListWho(fp);
    } else {
        for (int i=0; i < MAX_CONNECTIONS; i++)
            if (connArray[i])
                cnt += connArray[i]->ListWho(fp);
    }

    return(cnt);
}

//**************************************************************************
