/*****************************************************************************/
//    Term.h
/*****************************************************************************/

#ifndef _Term_H_
#define _Term_H_

#include "BindEnv.h"
#include "Arg.h"
#include "VarArg.h"
#include "NumArg.h"
#include "Assert.h"
#include "CArg.h"
/*****************************************************************************/
//
// A Term is implemented using a variant of structure sharing (though this 
// may change):
// The 'expr' field contains the static structure of the term,
// while 'bindenv' maps (some of) the variables in 'expr' to terms.
// Variables in these `terms' are not interpreted in BindEnv, unlike what is
// done in Boyer and Moore structure sharing.
//

class corTerm {
public:
  /* arg  is NULL if unbound ??? */
    corBindEnv *bindenv;     /* context for variables in u.arg */
    corCArg expr;
    inline corTerm() { expr.init(); bindenv = NULL; } 
    inline corTerm (corArg *arg1, corBindEnv *bindenv1 = NULL): expr(arg1) {
        bindenv = bindenv1;
    }

    corTerm(const corCArg& expr1, corBindEnv *bindenv1 = NULL): expr(expr1) {
      bindenv = bindenv1;
     }

    inline void printon(FILE*file) const {
      if (!expr.IsEmpty())
        expr.print(bindenv, file);
      else fputc('_', file);
    }

    // Assuming 'expr' is a VarArg*, get its bindings.
    inline corTerm dereference_var() const ; 
    inline Variable varnum() 
    //{ return (((VarArg*)expr)->var) ;}
     { return expr.VarArgP()->var ;}
    corTerm dereference();
    inline int hash() const { return expr.hash(bindenv); }
    inline arg_kind kindof() { return expr.kindof(); }
    inline int isConstant()
    { return (expr.isConstant());}
    ~corTerm(){};
};

extern corTerm corGlobalNewt;
extern corTerm corNullTerm ;

/*****************************************************************************/
//    Inline functions defined here.  They cannot be defined earlier since
//    they need to know the size of classes defined later.

/*****************************************************************************/
inline corTerm corTerm::dereference_var() const {
    ASSERT(expr.kindof()==COR_VARIABLE);
    return bindenv->lookup(((corVarArg*)expr.other)->var); 
}

/*****************************************************************************/
//    The macro ARITH_DEREFERENCE_TERM overwrites its argument.
//    Use it carefully.  It takes as argument a term, and if it is an arithmetic
//    expression it is simplified.


inline void corarith_dereference_term( corTerm &term) {
  corTermLink *renamed_vars = NULL;
  corTerm newt;                              
  newt.expr = (term).expr.simplify((term).bindenv, renamed_vars, 
                   (term).bindenv, NULL);      
  newt.bindenv = NULL;                            
  if (! newt.expr.IsEmpty() )                
    (term) = newt;                        
 }

/*****************************************************************************/
//    The macro FULL_DEREFERENCE_TERM overwrites its argument.  
//    Use it carefully.
//    It takes as argument a term, and dereferences the term fully, till the 
//    term is either a free variable or a non-variable structure or a constant.
//    Note that this merely chases down chains of variable bindings. It does NOT
//    go into structures.
#define FULL_DEREFERENCE_TERM(term) corfull_dereference_term(term)

inline void  corfull_dereference_term(corTerm &term) {
  if (((term).expr.IsEmpty())==0) {              
    arg_kind temp_kind = (term).expr.kindof();
    while ((term).bindenv    
       && (temp_kind == COR_VARIABLE)) {             
      corGlobalNewt =                                    
    (term).bindenv->lookup(term.expr.VarArgP()->var); 
      if (!(corGlobalNewt.expr.IsEmpty()))                 
    (term) = corGlobalNewt;                         
      else break;                                  
      temp_kind = (term).expr.kindof();                   
     }                                 
    if (temp_kind == COR_NUM_CONST &&                   
    term.expr.cNum_kindof() == COR_ARITH_EXPR) 
      corarith_dereference_term(term);
       }                               
 }


/*****************************************************************************/
//    The macro DEEP_DEREFERENCE_TERM overwrites its argument.  
//    Use it carefully.
//    It takes as argument a term, and deeply dereferences the term, by replacing
//    all bound variables by the appropriate constant.


inline void cordeep_dereference_term(corTerm &term) {     
  corfull_dereference_term(term);    
  if (!term.isConstant() && corisGroundHashValue(term.hash())) {
    corCArg new_arg;                        
    corCopyArg(new_arg, term.expr, term.bindenv);    
    term.expr = new_arg;    
    term.bindenv = NULL;
   }                        
 } 

/*****************************************************************************/
//    ISFREE_TERM - returns 1 if argument  is a free variable, 0 otherwise 

inline  int corisfree_term(corTerm *t) {
  return ((t->expr.kindof()==COR_VARIABLE) &&
    (t->bindenv->lookup(t->expr.VarArgP()->var).expr.IsEmpty()));
 }

/*****************************************************************************/

#define ceil_div(size, incr) ((size+incr-1)/incr)


/*****************************************************************************/
inline corTerm corVarArg::dereference_var(corBindEnv* env)
{
    // Does a one level dereference
    ASSERT(env!=NULL);
    return env->lookup(var);
}

/*****************************************************************************/
inline corTerm corVarArg::dereference(corBindEnv* env)
{
    // Does a full dereference
    ASSERT(env!=NULL);
    return env->lookup(var).dereference();
}

/*****************************************************************************/
// TermLink is used by CopyArgs() to re-number VarArgs.

struct corTermLink {
    struct corTermLink *next;        // Next TermLink in list
    Variable old_var;
    corBindEnv  *old_bindenv;
    struct corVarArg *new_arg;        // ... and yield this new VarArg instead
};

/*****************************************************************************/

#endif _Term_H_


