// *************************************************************************
//	SipOrder.C
// *************************************************************************

#include "SipOrder.h"

#include "Rule.h"
#include "Assert.h"
#include "Sips.h"

/*****************************************************************************/

static int torder[MAX_RHSPREDS];

static void toporder(corClause *rule, int torder[]);

/*****************************************************************************/
/*
	graph(i,j) = 1 if there is a sip arc from pred[i] to pred[j] 
		(i = 0 implies the head of the rule )
*/

static void toporder(corClause *rule, int torder[])
{
	static char	graph[MAX_RHSPREDS+1][MAX_RHSPREDS+1];
	static int countarr[MAX_RHSPREDS+1];
	static int noarcs[MAX_RHSPREDS+1];
	int	i, j, k, count;
	struct corsip *s;
	int corb;

	ASSERT(rule->num_preds <= MAX_RHSPREDS);
	for(i=0; i <= rule->num_preds; i++)
		countarr[i] = 0;

	//	Initialize graph.
	for (i=0; i<=rule->num_preds; ++i) {
	    for (j=0; j<=rule->num_preds; ++j)
			graph[i][j] = 0;
		noarcs[i] = 1;
	}

	//	Construct graph from sips ans set up count array.
	for (s = rule->sips->sip_list; s; s = s->next) {
		for (corb = firstpredinsip(&s->tail); corb>=0; corb = nextpredinsip(&s->tail, corb))
		{
			graph[corb][s->head] = 1;
			noarcs[corb] = noarcs[s->head] = 0;
		}
	}

   	for (i=0; i<=rule->num_preds; i++) {
   	    for (j=0; j<=rule->num_preds; j++) {
			if ( graph[i][j] ) countarr[j]++;
		}
	}

	//	Force all preds with no incoming and no outgoing sip arcs to the 
	//	end of the rule by giving them a count of one.
	for (i=1; i <= rule->num_preds; i++)
		if (noarcs[i]) countarr[i] = 1;

	count = 0;
	for (k=0; k<=rule->num_preds; ++k) {
		for (i=0; i<=rule->num_preds; ++i) {
			if (countarr[i] == 0) {
				//	Mark as done.
				countarr[i] = -1;
				if (i > 0) torder[i-1] = count++;
				//	Bump down the pred numbering by 1 for torder to start
				//	body pred numbering at 0.
				for (j=1; j <= rule->num_preds; j++) {
					if (graph[i][j]) {
						countarr[j]--;
						ASSERT( countarr[j] >= 0 );
					}
				}
			}
		}
	}

	//	Set up rule->last_part properly, and make sure the order of
	//	predicates in the rule body is such that anything beyond last_part
	//	does not participate in any sip.

	rule->last_part = count -1; 
	//	Now tack on preds that had no sip arcs entering or leaving.
	for (i=1; i <= rule->num_preds; i++) {
	    if (noarcs[i]) {
	    	ASSERT(countarr[i] == 1);
	    	torder[i-1] = count++;
		}
	}


#	ifdef DEBUG
		if (0) {
			corprint_rule(rule, stderr);
			printf("torder array:");
			for (i = 0; i < rule->num_preds; i++)
				printf("%d ", i );
			printf("\n");
		}
#	endif
}

/*****************************************************************************/
/*
	Reorder the right-hand sides of the adorned system so the predicate
	occurrences follow the order induced by the sips.
 */

void siporder(struct corcollection *ad_rules)
{
	int i,corb;
	struct corsip* s;
	struct corsips* sp;
	corLiteral* newpreds[MAX_RHSPREDS];

	FOR_EACH_RULE(adrule, ad_rules) {
		/*
		printf("Here is the original rule\n");
		print_rule(adrule, stderr);
		printf("\n");
		*/

		toporder(adrule, torder);

		//	Rearrange right-hand side.
		for (i=0; i<adrule->num_preds; ++i) {
			newpreds[torder[i]] = adrule->preds[i];
		}
		for (i=0; i<adrule->num_preds; ++i) {
			adrule->preds[i] = newpreds[i];
		}

		//	Fix sips to reflect rearrangement.
		for (sp = adrule->sips; sp; sp = sp->next) {
			for (s = sp->sip_list; s; s = s->next) {
				corBitVector bitvec(s->tail.len());
				for (corb = firstpredinsip(&s->tail); corb>=0;
					corb = nextpredinsip(&s->tail, corb))
				{
					if (corb == 0) {
						//	Head predicate
						bitvec.set(0);
					}
					else {
						//	Add 1 since 0 is head
						bitvec.set(torder[corb-1]+1); 
					}
				}
				s->tail = bitvec;
				s->head = torder[s->head-1]+1;
				
			}
		}

#	ifdef DEBUG
		if (0) {
			printf("Modified original rule:\n");
			corprint_rule(adrule, stderr);
			printf("\n");
		}
#	endif
	} END_EACH_RULE
}

/*****************************************************************************/
/*
	Find a sip for a predicate in an adorned rule.  Non-matching sips have
	already been discarded.
*/

struct corsip* findadsip(corClause* adrule, int p)
{
	struct corsip *s;

	for (s=adrule->sips->sip_list; s; s=s->next) {
		//	In sip, body pred numbering starts from 1.
		if (p == s->head-1) break; 
	}
	return(s);
}

/*****************************************************************************/

